"""Resource Server utils functions."""

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from joserfc.jwk import JWKRegistry


def import_private_key_from_settings():
    """
    Import the private key used by the resource server when interacting with the OIDC provider.

    This private key is crucial; its public components are exposed in the JWK endpoints,
    while its private component is used for decrypting the introspection token retrieved
    from the OIDC provider.

    By default, we recommend using RSAKey for asymmetric encryption,
    known for its strong security features.

    Note:
        - The function requires the 'OIDC_RS_PRIVATE_KEY_STR' setting to be configured.
        - The 'OIDC_RS_ENCRYPTION_KEY_TYPE' and 'OIDC_RS_ENCRYPTION_ALGO' settings can be customized
          based on the chosen key type.

    Raises:
        ImproperlyConfigured: If the private key setting is missing, empty, or incorrect.

    Returns:
        joserfc.jwk.JWK: The imported private key as a JWK object.

    """
    private_key_str = getattr(settings, "OIDC_RS_PRIVATE_KEY_STR", None)
    if not private_key_str:
        raise ImproperlyConfigured("OIDC_RS_PRIVATE_KEY_STR setting is missing or empty.")

    private_key_pem = private_key_str.encode()

    try:
        private_key = JWKRegistry.import_key(
            private_key_pem,
            key_type=settings.OIDC_RS_ENCRYPTION_KEY_TYPE,
            parameters={"alg": settings.OIDC_RS_ENCRYPTION_ALGO, "use": "enc"},
        )
    except ValueError as err:
        raise ImproperlyConfigured("OIDC_RS_PRIVATE_KEY_STR setting is wrong.") from err

    return private_key



"""Mixins for resource server views."""

from rest_framework import exceptions as drf_exceptions

from .authentication import ResourceServerAuthentication


class ResourceServerMixin:
    """
    Mixin for resource server views:
     - Restrict the authentication class to ResourceServerAuthentication.
     - Adds the Service Provider ID to the serializer context.
     - Fetch the Service Provider ID from the OIDC introspected token stored
       in the request.

    This Mixin *must* be used in every view that should act as a resource server.
    """

    authentication_classes = [ResourceServerAuthentication]

    def get_serializer_context(self):
        """Extra context provided to the serializer class."""
        context = super().get_serializer_context()

        # When used as a resource server, we need to know the audience to automatically:
        # - add the Service Provider to the team "scope" on creation
        context["from_service_provider_audience"] = self._get_service_provider_audience()

        return context

    def _get_service_provider_audience(self):
        """Return the audience of the Service Provider from the OIDC introspected token."""
        if not isinstance(self.request.successful_authenticator, ResourceServerAuthentication):
            # We could check request.resource_server_token_audience here, but it's
            # more explicit to check the authenticator type and assert the attribute
            # existence.
            return None

        # When used as a resource server, the request has a token audience
        service_provider_audience = self.request.resource_server_token_audience

        if not service_provider_audience:  # should not happen
            raise drf_exceptions.AuthenticationFailed("Resource server token audience not found in request")

        return service_provider_audience



"""Resource Server Authentication."""

import base64
import binascii
import contextlib
import logging
from functools import cache

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from mozilla_django_oidc.contrib.drf import OIDCAuthentication

from .backend import ResourceServerImproperlyConfiguredBackend

logger = logging.getLogger(__name__)


@cache
def get_resource_server_backend():
    """Return the resource server backend class based on the settings."""
    return import_string(settings.OIDC_RS_BACKEND_CLASS)


class ResourceServerAuthentication(OIDCAuthentication):
    """Authenticate clients using the token received from the authorization server."""

    def __init__(self):
        """Require authentication to be configured in order to instantiate."""
        super().__init__()

        try:
            self.backend = get_resource_server_backend()()

        except ImproperlyConfigured as err:
            message = "Resource Server authentication is disabled"
            logger.debug("%s. Exception: %s", message, err)
            self.backend = ResourceServerImproperlyConfiguredBackend()

    def get_access_token(self, request):
        """
        Retrieve and decode the access token from the request.

        This method overrides the 'get_access_token' method from the parent class,
        to support service providers that would base64 encode the bearer token.
        """
        access_token = super().get_access_token(request)

        with contextlib.suppress(binascii.Error, TypeError, UnicodeDecodeError):
            access_token = base64.b64decode(access_token, validate=True).decode("utf-8")

        return access_token

    def authenticate(self, request):
        """
        Authenticate the request and return a tuple of (user, token) or None.

        We override the 'authenticate' method from the parent class to store
        the introspected token audience inside the request.
        """
        result = super().authenticate(request)  # Might raise AuthenticationFailed

        if result is None:  # Case when there is no access token
            return None

        # Note: at this stage, the request is a "drf_request" object
        request.resource_server_token_audience = self.backend.token_origin_audience

        return result



"""Authentication URLs for the OIDC backend."""

from django.urls import path
from mozilla_django_oidc.urls import urlpatterns as mozilla_oidc_urls

from .views import OIDCLogoutCallbackView, OIDCLogoutView

urlpatterns = [
    # Override the default 'logout/' path from Mozilla Django OIDC with our custom view.
    path("logout/", OIDCLogoutView.as_view(), name="oidc_logout_custom"),
    path(
        "logout-callback/",
        OIDCLogoutCallbackView.as_view(),
        name="oidc_logout_callback",
    ),
    *mozilla_oidc_urls,
]



"""Authentication Views for the OIDC authentication backend."""

import copy
from urllib.parse import urlencode

from django.contrib import auth
from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils import crypto
from mozilla_django_oidc.utils import (
    absolutify,
)
from mozilla_django_oidc.views import (
    OIDCAuthenticationCallbackView as MozillaOIDCAuthenticationCallbackView,
)
from mozilla_django_oidc.views import (
    OIDCAuthenticationRequestView as MozillaOIDCAuthenticationRequestView,
)
from mozilla_django_oidc.views import (
    OIDCLogoutView as MozillaOIDCOIDCLogoutView,
)


class OIDCLogoutView(MozillaOIDCOIDCLogoutView):
    """
    Custom logout view for handling OpenID Connect (OIDC) logout flow.

    Adds support for handling logout callbacks from the identity provider (OP)
    by initiating the logout flow if the user has an active session.

    The Django session is retained during the logout process to persist the 'state' OIDC parameter.
    This parameter is crucial for maintaining the integrity of the logout flow between this call
    and the subsequent callback.
    """

    @staticmethod
    def persist_state(request, state):
        """
        Persist the given 'state' parameter in the session's 'oidc_states' dictionary.

        This method is used to store the OIDC state parameter in the session, according to the
        structure expected by Mozilla Django OIDC's 'add_state_and_verifier_and_nonce_to_session'
        utility function.
        """
        if "oidc_states" not in request.session or not isinstance(request.session["oidc_states"], dict):
            request.session["oidc_states"] = {}

        request.session["oidc_states"][state] = {}
        request.session.save()

    def construct_oidc_logout_url(self, request):
        """
        Create the redirect URL for interfacing with the OIDC provider.

        Retrieves the necessary parameters from the session and constructs the URL
        required to initiate logout with the OpenID Connect provider.

        If no ID token is found in the session, the logout flow will not be initiated,
        and the method will return the default redirect URL.

        The 'state' parameter is generated randomly and persisted in the session to ensure
        its integrity during the subsequent callback.
        """
        oidc_logout_endpoint = self.get_settings("OIDC_OP_LOGOUT_ENDPOINT")

        if not oidc_logout_endpoint:
            return self.redirect_url

        reverse_url = reverse("oidc_logout_callback")
        id_token = request.session.get("oidc_id_token", None)

        if not id_token:
            return self.redirect_url

        query = {
            "id_token_hint": id_token,
            "state": crypto.get_random_string(self.get_settings("OIDC_STATE_SIZE", 32)),
            "post_logout_redirect_uri": absolutify(request, reverse_url),
        }

        self.persist_state(request, query["state"])

        return f"{oidc_logout_endpoint}?{urlencode(query)}"

    def post(self, request):
        """
        Handle user logout.

        If the user is not authenticated, redirects to the default logout URL.
        Otherwise, constructs the OIDC logout URL and redirects the user to start
        the logout process.

        If the user is redirected to the default logout URL, ensure her Django session
        is terminated.
        """
        logout_url = self.redirect_url

        if request.user.is_authenticated:
            logout_url = self.construct_oidc_logout_url(request)

        # If the user is not redirected to the OIDC provider, ensure logout
        if logout_url == self.redirect_url:
            auth.logout(request)

        return HttpResponseRedirect(logout_url)


class OIDCLogoutCallbackView(MozillaOIDCOIDCLogoutView):
    """
    Custom view for handling the logout callback from the OpenID Connect (OIDC) provider.

    Handles the callback after logout from the identity provider (OP).
    Verifies the state parameter and performs necessary logout actions.

    The Django session is maintained during the logout process to ensure the integrity
    of the logout flow initiated in the previous step.
    """

    http_method_names = ["get"]

    def get(self, request):
        """
        Handle the logout callback.

        If the user is not authenticated, redirects to the default logout URL.
        Otherwise, verifies the state parameter and performs necessary logout actions.
        """
        if not request.user.is_authenticated:
            return HttpResponseRedirect(self.redirect_url)

        state = request.GET.get("state")

        if state not in request.session.get("oidc_states", {}):
            msg = "OIDC callback state not found in session `oidc_states`!"
            raise SuspiciousOperation(msg)

        del request.session["oidc_states"][state]
        request.session.save()

        auth.logout(request)

        return HttpResponseRedirect(self.redirect_url)


class OIDCAuthenticationCallbackView(MozillaOIDCAuthenticationCallbackView):
    """Custom callback view for handling the silent login flow."""

    @property
    def failure_url(self):
        """
        Override the failure URL property to handle silent login flow.

        A silent login failure (e.g., no active user session) should not be
        considered as an authentication failure.
        """
        if self.request.session.get("silent", None):
            del self.request.session["silent"]
            self.request.session.save()
            return self.success_url
        return super().failure_url


class OIDCAuthenticationRequestView(MozillaOIDCAuthenticationRequestView):
    """Custom authentication view for handling the silent login flow."""

    def get_extra_params(self, request):
        """
        Handle 'prompt' extra parameter for the silent login flow.

        This extra parameter is necessary to distinguish between a standard
        authentication flow and the silent login flow.
        """
        extra_params = self.get_settings("OIDC_AUTH_REQUEST_EXTRA_PARAMS", None)
        if extra_params is None:
            extra_params = {}
        if request.GET.get("silent", "").lower() == "true":
            extra_params = copy.deepcopy(extra_params)
            extra_params.update({"prompt": "none"})
            request.session["silent"] = True
            request.session.save()
        return extra_params



"""Module containing the tasks for the JCOP backend."""

import logging

import requests
from celery import shared_task

from .. import MalwareDetectionHandler
from ..exceptions import MalwareDetectionInvalidAuthenticationError

logger = logging.getLogger(__name__)


@shared_task(
    bind=True,
    default_retry_delay=3,
    max_retries=100,
    dont_autoretry_for=(MalwareDetectionInvalidAuthenticationError,),
)
def analyse_file_async(
    self,
    file_path: str,
    file_hash: str | None = None,
    **kwargs,
) -> None:
    """Task starting analysis process for a file."""
    backend = MalwareDetectionHandler()()  # JCOPBackend
    try:
        should_retry = backend.check_analysis(file_path, file_hash=file_hash, **kwargs)
    except requests.exceptions.RequestException as exc:
        if self.request.retries >= self.max_retries:
            backend.failed_analysis(file_path, "Max retries fetching results exceeded")
            return
        self.retry(exc=exc)

    if should_retry:
        self.retry()


@shared_task(bind=True, max_retries=6, dont_autoretry_for=(MalwareDetectionInvalidAuthenticationError,))
def trigger_new_analysis(
    self,
    file_path: str,
    **kwargs,
) -> None:
    """Trigger a new analysis for a file."""
    backend = MalwareDetectionHandler()()  # JCOPBackend
    try:
        backend.trigger_new_analysis(file_path, **kwargs)
    except requests.exceptions.RequestException as exc:
        if self.request.retries >= self.max_retries:
            backend.failed_analysis(file_path, "Max retries triggering new analysis exceeded")
            return
        self.retry(exc=exc)
        return
    except TimeoutError:
        if self.request.retries >= self.max_retries:
            backend.failed_analysis(file_path, "Max retries triggering new analysis exceeded")
            return
        self.retry(exc=TimeoutError())
        return
